from microcosm_flask.formatting.base import BaseFormatter


class HTMLFormatter(BaseFormatter):

    CONTENT_TYPE = "text/html"

    @property
    def content_type(self):
        return HTMLFormatter.CONTENT_TYPE

    def format(self, response_data):
        return response_data



from os import makedirs
from os.path import exists, expanduser

from werkzeug.contrib.profiler import ProfilerMiddleware


def default_profile_dir(name):
    return expanduser("~/.{name}/profile".format(name=name))


def enable_profiling(graph):
    profile_dir = graph.config.flask.profile_dir or default_profile_dir(name=graph.metadata.name)
    if not exists(profile_dir):
        makedirs(profile_dir)

    graph.app.config['PROFILE'] = True
    graph.app.wsgi_app = ProfilerMiddleware(
        graph.app.wsgi_app,
        profile_dir=profile_dir,
    )

    graph.app.logger.info("*** Profiling is ON, Will save profiling data to directory: {}".format(profile_dir))



"""
Topological sort.

"""
from collections import defaultdict
from logging import getLogger

TEMPORARY = object()
VISITED = object()


logger = getLogger("sync.toposort")


def iter_parents(resource):
    for relation, links in resource["_links"].items():
        if relation == "self" or relation.startswith("child:"):
            continue
        if isinstance(links, list):
            for link in links:
                yield link["href"]
        else:
            yield links["href"]


def toposorted(inputs):
    """
    Perform a topological sort on the input (href, resource) tuples.

    Uses a DFS.

    """
    sorted_items = []

    visited = {}

    nodes = {
        href: resource
        for href, resource in inputs
    }

    children = defaultdict(list)
    for href, resource in nodes.items():
        for parent_href in iter_parents(resource):
            children[parent_href].append(href)

    def visit(href, resource):
        if visited.get(href) == VISITED:
            return
        if visited.get(href) == TEMPORARY:
            raise Exception("Found cycle at {}".format(href))

        visited[href] = TEMPORARY
        for child_href in children[href]:
            visit(child_href, nodes[child_href])
        visited[href] = VISITED

        sorted_items.append((href, resource))

    logger.info("Toposorting {} resources".format(len(nodes)))

    for href, resource in nodes.items():
        visit(href, resource)

    return reversed(sorted_items)



"""
Serve build information.

"""
from microcosm.api import defaults
from microcosm_flask.audit import skip_logging
from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import make_response
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation


class BuildInfo:

    def __init__(self, build_num, sha1):
        self.build_num = build_num
        self.sha1 = sha1

    def to_dict(self):
        return dict(
            build_num=self.build_num,
            sha1=self.sha1,
        )

    @classmethod
    def from_config(cls, config):
        return cls(
            build_num=config.build_info_convention.build_num,
            sha1=config.build_info_convention.sha1,
        )

    @staticmethod
    def check_build_num(graph):
        build_info = BuildInfo.from_config(graph.config)
        return build_info.build_num or "undefined"

    @staticmethod
    def check_sha1(graph):
        build_info = BuildInfo.from_config(graph.config)
        return build_info.sha1 or "undefined"


class BuildInfoConvention(Convention):

    def __init__(self, graph):
        super(BuildInfoConvention, self).__init__(graph)
        self.build_info = BuildInfo.from_config(graph.config)

    def configure_retrieve(self, ns, definition):

        @self.add_route(ns.singleton_path, Operation.Retrieve, ns)
        @skip_logging
        def build_info():
            response_data = self.build_info.to_dict()
            return make_response(response_data)


@defaults(
    build_num=None,
    sha1=None,
)
def configure_build_info(graph):
    """
    Configure the build info endpoint.

    """
    ns = Namespace(
        subject=BuildInfo,
    )

    convention = BuildInfoConvention(graph)
    convention.configure(ns, retrieve=tuple())
    return convention.build_info



"""
Conventions for file upload.

"""
from contextlib import contextmanager
from functools import wraps
from os.path import join
from shutil import rmtree
from tempfile import mkdtemp

from flask import request
from marshmallow import Schema
from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import (
    dump_response_data,
    load_query_string_data,
    merge_data,
)
from microcosm_flask.conventions.registry import qs, response
from microcosm_flask.operations import Operation
from werkzeug.exceptions import BadRequest
from werkzeug.utils import secure_filename


try:
    # Python 2
    from contextlib import nested
except ImportError:
    # Python3
    from contextlib import ExitStack

    @contextmanager
    def nested(*contexts):
        """
        Reimplementation of nested in python 3.
        """
        with ExitStack() as stack:
            results = [
                stack.enter_context(context)
                for context in contexts
            ]
            yield results


@contextmanager
def temporary_upload(name, fileobj):
    """
    Upload a file to a temporary location.

    Flask will not load sufficiently large files into memory, so it
    makes sense to always load files into a temporary directory.

    """
    tempdir = mkdtemp()
    filename = secure_filename(fileobj.filename)
    filepath = join(tempdir, filename)
    fileobj.save(filepath)
    try:
        yield name, filepath, fileobj.filename
    finally:
        rmtree(tempdir)


class UploadConvention(Convention):

    def __init__(self, graph, exclude_func=None):
        Convention.__init__(self, graph)

        self.exclude_func = exclude_func or (lambda name, fileobj: False)

    def create_upload_func(self, ns, definition, path, operation):
        request_schema = definition.request_schema or Schema()
        response_schema = definition.response_schema or Schema()

        @self.add_route(path, operation, ns)
        @wraps(definition.func)
        def upload(**path_data):
            merged_data = merge_data(request.args.to_dict(), request.form.to_dict())
            request_data = load_query_string_data(request_schema, merged_data)

            if not request.files:
                raise BadRequest("No files were uploaded")

            uploads = [
                temporary_upload(name, fileobj)
                for name, fileobj
                in request.files.items()
                if not self.exclude_func(name, fileobj)
            ]
            with nested(*uploads) as files:
                response_data = definition.func(files, **merge_data(path_data, request_data))
                if response_data is None:
                    return "", 204

            return dump_response_data(response_schema, response_data, operation.value.default_code)

        if definition.request_schema:
            upload = qs(definition.request_schema)(upload)
        if definition.response_schema:
            upload = response(definition.response_schema)(upload)
        return upload

    def configure_upload(self, ns, definition):
        """
        Register an upload endpoint.

        The definition's func should be an upload function, which must:
        - accept kwargs for path data and query string parameters
        - accept a list of tuples of the form (formname, tempfilepath, filename)
        - optionally return a resource

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        upload = self.create_upload_func(ns, definition, ns.collection_path, Operation.Upload)
        upload.__doc__ = "Upload a {}".format(ns.subject_name)

    def configure_uploadfor(self, ns, definition):
        """
        Register an upload-for relation endpoint.

        The definition's func should be an upload function, which must:
        - accept kwargs for path data and query string parameters
        - accept a list of tuples of the form (formname, tempfilepath, filename)
        - optionall return a resource

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        upload_for = self.create_upload_func(ns, definition, ns.relation_path, Operation.UploadFor)
        upload_for.__doc__ = "Upload a {} for a {}".format(ns.subject_name, ns.object_name)


def configure_upload(graph, ns, mappings, exclude_func=None):
    """
    Register Upload endpoints for a resource object.

    """
    convention = UploadConvention(graph, exclude_func)
    convention.configure(ns, mappings)



"""
CSV response formatting.

"""
from csv import writer, QUOTE_MINIMAL
from io import StringIO

from microcosm_flask.formatting.base import BaseFormatter


class CSVFormatter(BaseFormatter):

    CONTENT_TYPE = "text/csv"

    @property
    def content_type(self):
        return CSVFormatter.CONTENT_TYPE

    def build_headers(self, headers, **kwargs):
        # TODO: pass in optional filename
        filename = "response.csv"
        headers["Content-Disposition"] = "attachment; filename=\"{}\"".format(filename)

        return headers

    def get_column_names(self, list_response_data):
        response_fields = list(list_response_data[0].keys())

        column_order = getattr(self.response_schema, "csv_column_order", None)
        if column_order is None:
            # We should still be able to return a CSV even if no column order has been specified
            column_names = response_fields
        else:
            column_names = self.response_schema.csv_column_order

            # The column order be only partially specified
            column_names.extend([field_name for field_name in response_fields if field_name not in column_names])

        return column_names

    def format(self, response_data):
        """
        Make Flask `Response` object, with data returned as a generator for the CSV content
        The CSV is built from JSON-like object (Python `dict` or list of `dicts`)

        """
        if "items" in response_data:
            list_response_data = response_data["items"]
        else:
            list_response_data = [response_data]

        write_column_names = type(list_response_data[0]) not in (tuple, list)

        output = StringIO()
        csv_writer = writer(output, quoting=QUOTE_MINIMAL)

        if write_column_names:
            column_names = self.get_column_names(list_response_data)
            csv_writer.writerow(column_names)

        for item in list_response_data:
            csv_writer.writerow(
                [item[column] for column in column_names] if write_column_names else list(item)
            )

        # Ideally we'd want to `yield` each line to stream the content
        # But something downstream seems to break streaming
        yield output.getvalue()



"""
Audit log control decorators.

"""
from functools import wraps

from flask import g


def hide(*keys):
    """
    Hide a set of request and/or response fields from logs.

    Example:

        @hide("id")
        def create_foo():
            return Foo(id=uuid4())

    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            g.hide_request_fields = keys
            g.hide_response_fields = keys
            return func(*args, **kwargs)
        return wrapper
    return decorator


def show_as(**mappings):
    """
    Show a set of request and/or response fields in logs using a different key.

    Example:

        @show_as(id="foo_id")
        def create_foo():
            return Foo(id=uuid4())

    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            g.show_request_fields = mappings
            g.show_response_fields = mappings
            return func(*args, **kwargs)
        return wrapper
    return decorator



"""
Naming conventions for Swagger definitions.

Intended to play nice with generated code (e.g. with Bravado)

"""

from inflection import camelize, pluralize


def operation_name(operation, ns):
    """
    Convert an operation, obj(s) pair into a swagger operation id.

    For compatability with Bravado, we want to use underscores instead of dots and
    verb-friendly names. Example:

        foo.retrieve       => client.foo.retrieve()
        foo.search_for.bar => client.foo.search_for_bars()

    """
    verb = operation.value.name
    if ns.object_:
        return "{}_{}".format(verb, pluralize(ns.object_name))
    else:
        return verb


def type_name(name):
    """
    Convert an internal name into a swagger type name.

    For example:

        foo_bar => FooBar

    """
    if name.endswith("_schema"):
        name = name[:-7]
    return camelize(name)
